@@ -16,6 +16,7 @@
  * @EXPR_INVALID:	uninitialized type, should not happen
  * @EXPR_VERDICT:	nftables verdict expression
  * @EXPR_SYMBOL:	unparsed symbol
+ * @EXPR_VARIABLE:	variable
  * @EXPR_VALUE:		literal numeric or string expression
  * @EXPR_PREFIX:	prefixed expression
  * @EXPR_RANGE:		literal range
@@ -41,6 +42,7 @@ enum expr_types {
 	EXPR_INVALID,
 	EXPR_VERDICT,
 	EXPR_SYMBOL,
+	EXPR_VARIABLE,
 	EXPR_VALUE,
 	EXPR_PREFIX,
 	EXPR_RANGE,
@@ -83,12 +85,6 @@ enum ops {
 	OP_GT,
 	OP_LTE,
 	OP_GTE,
-	/* Range comparison */
-	OP_RANGE,
-	/* Flag comparison */
-	OP_FLAGCMP,
-	/* Set lookup */
-	OP_LOOKUP,
 	__OP_MAX
 };
 #define OP_MAX		(__OP_MAX - 1)
@@ -97,7 +93,6 @@ extern const char *expr_op_symbols[];
 
 enum symbol_types {
 	SYMBOL_VALUE,
-	SYMBOL_DEFINE,
 	SYMBOL_SET,
 };
 
@@ -225,6 +220,10 @@ struct expr {
 			const char		*identifier;
 			enum symbol_types	symtype;
 		};
+		struct {
+			/* EXPR_VARIABLE */
+			struct symbol		*sym;
+		};
 		struct {
 			/* EXPR_VERDICT */
 			int			verdict;
@@ -387,6 +386,9 @@ static inline void symbol_expr_set_type(struct expr *expr,
 		expr->dtype = dtype;
 }
 
+struct expr *variable_expr_alloc(const struct location *loc,
+				 struct scope *scope, struct symbol *sym);
+
 extern struct expr *constant_expr_alloc(const struct location *loc,
 					const struct datatype *dtype,
 					enum byteorder byteorder,
@@ -408,6 +410,8 @@ extern struct expr *prefix_expr_alloc(const struct location *loc,
 extern struct expr *range_expr_alloc(const struct location *loc,
 				     struct expr *low, struct expr *high);
 
+extern struct expr *compound_expr_alloc(const struct location *loc,
+					const struct expr_ops *ops);
 extern void compound_expr_add(struct expr *compound, struct expr *expr);
 extern void compound_expr_remove(struct expr *compound, struct expr *expr);
 extern void list_expr_sort(struct list_head *head);
@@ -423,6 +427,11 @@ extern int set_to_intervals(struct list_head *msgs, struct set *set,
 			    unsigned int debug_mask, bool merge);
 extern void interval_map_decompose(struct expr *set);
 
+extern struct expr *get_set_intervals(const struct set *set,
+				      const struct expr *init);
+struct table;
+extern void get_set_decompose(struct table *table, struct set *set);
+
 extern struct expr *mapping_expr_alloc(const struct location *loc,
 				       struct expr *from, struct expr *to);
 extern struct expr *map_expr_alloc(const struct location *loc,
